#!/usr/bin/perl -w
#Program to do gradient-descent best-fitting
#Author: Peter Behroozi
#Creation Date: 3/24/08
#Compilation instructions (requires gcc & perl): make

my $wf = 10;
my $dims = $ARGV[0] || $wf;
my $d = 4.7;
my @c = map { 0 } (1..$wf);
my @dc = map { 0.05 } (1..$wf);
$c[0] = 1;

my ($err, $last_err)=(1e10, 1e11);
$err = chi2_error();

while (abs($last_err-$err)>0.01) {
    $last_err = $err;

    for (0..($dims-1)) { improve_fit(\$err, \$c[$_], \$dc[$_]); }
    print("Now at $d @c.\n");
}
$err = chi2_error();
print "Final energy: $err eV\n";
print "Final distance: ", $d*0.264588544," Angstrom\n";

sub improve_fit {
    my ($err, $param, $amount) = @_;
    my $last_err = $$err;
    $$param += $$amount/5;
    my $err_r = chi2_error();
    $$param -= $$amount/2.5;
    my $err_l = chi2_error();
    $$param += $$amount/5;
    my $d = minimum_delta($err_l, $$err, $err_r, $$amount/5);

    if (abs($d)>$$amount) { $d *= $$amount/abs($d); }
    my $old_param = $$param;
    $$param += $d;
    $$amount/=2;

    $$err = chi2_error();
    if ($err_r<$$err) { $$param = $old_param+$$amount/2.5; $$err=$err_r; }
    if ($err_l<$$err) { $$param = $old_param-$$amount/2.5; $$err=$err_l; }
    if ($last_err<$$err) { $$param = $old_param; $$err = $last_err; }
}



# Given chi2_l, chi2, chi2_r, ds, finds the distance to the
# minimum of the error chi_2
sub minimum_delta {
    my ($c1, $c2, $c3, $ds) = @_;
    my $dc1 = ($c2-$c1)/$ds;
    my $dc2 = ($c3 - $c2)/$ds;
    my $ddc = ($dc2 - $dc1)/$ds;
    my $dc = 0.5*($dc1 + $dc2);
    if ($c2>$c3 && $c2>$c1) { return (($c1<$c3) ? -$ds : $ds); }
    return $ddc ? (-$dc/$ddc) : 0;
}


sub _chi2_err {
    my $en =  `./hf @_`;
    chomp($en);
    return $en;
}

sub chi2_error
{
    #find best fit for distance d
    $d = 4.5;
    my $dd = 0.1;
    my $last_err = 1e10;
    my $err = _chi2_err($d, @c);

    while (abs($last_err-$err)>0.01) {
	$last_err = $err;
	my $err_r = _chi2_err($d+$dd, @c);
	my $err_l = _chi2_err($d-$dd, @c);
	
	# Calculate error minimum
	my $de = minimum_delta($err_l, $err, $err_r, $dd);
	if (abs($de)>$dd*5) { $de *= $dd*5/abs($de); }
	my $old_d = $d;
	$d+=$de;
	$dd/=2;

	# Calculate new error---revert to old values if those are better.
	$err = _chi2_err($d, @c);
	if ($err_r<$err) { $d = $old_d+$dd*2; $err=$err_r; }
	if ($err_l<$err) { $d = $old_d-$dd*2; $err=$err_l; }
	if ($last_err<$err) { $d = $old_d; $err=$last_err; }
    }
    $err = _chi2_err($d, @c);

    print("Energy $d @c: $err\n");
    return $err;
}
